/**
Portions Copyright (c) Microsoft Corporation
 */

/**
 * Definitions for the SparkMonitor singleton object.
 * @module SparkMonitor
 */

import currentcell from './currentcell'; // Module to detect currently running cell. CodeCellModel interface
import CellMonitor from './CellMonitor'; // CellMonitor object constructor. CodeCell interface
const { v4: uuidv4 } = require('uuid');

export default class SparkMonitor {
    /**
     * SparkMonitor is the main singleton class that is responsible for managing CellMonitor instances for cells that run spark jobs.
     * It also delegates spark lifecycle events from the backend to corresponding CellMonitor.
     * @constructor
     * @param {NotebookPanel} nbPanel - The Jupyter NotebookPanel instance
     *
     */
    constructor(nbPanel) {
        console.log('SparkMonitor constructor');
        this.id = uuidv4();
        this.nbPanel = nbPanel;
        this.cellmonitors = {};
        this.comm = null;

        /** Data mapping jobs to cells for delegating further lifecycle events of a job. */
        this.data = {};
        this.appName = 'NULL';
        this.appId = 'NULL';
        this.app = 'NULL';
        this.totalCores = 0;
        this.numExecutors = 0;
        this.display_mode = 'shown'; // "shown" || "hidden"

        function registerCell(cell, nbPanel, that) {
            if (cell.type === 'code') {
                cell.metadata.changed.connect((_, data) => {
                    if (data.key === "execution" && data.type === "add") {
                        // set current cell when cell start
                        currentcell.cellExecuteStarted(null, {
                            'cell': cell
                        });
                    }
                })

                cell.outputs.list.changed.connect((_, data) => {
                    if (data.type === "remove") {
                        // clear cellMonitor when output cleared
                        that.removeCellMonitor(cell);
                    }
                })
            }
        }

        // register event handlers on current cells
        const cell_count = this.nbPanel.content.model.cells.length;
        for (let i = 0; i < cell_count; ++i) {
            let cell = this.nbPanel.content.model.cells.get(i);
            registerCell(cell, this.nbPanel, this);
        }

        // handle cell add and remove
        this.nbPanel.content.model.cells.changed.connect((_, data) => {
            // console.log('SparkMonitor on cell', data.type);
            if (data.type === 'remove') {
                // on cell removed, also clear cellMonitor and current cell
                for (let cell of data.oldValues) {
                    this.removeCellMonitor(cell);
                    currentcell.cellDeleted(null, {
                        cell
                    });
                }
            } else if (data.type === 'add') {
                // on cell added, register cell event handlers
                for (let cell of data.newValues) {
                    registerCell(cell, this.nbPanel, this)
                }
            }
        });

        // Fixes Reloading the browser
        this.startComm();

        // Fixes Restarting the Kernel
        // Fixes Changing the Kernel
        if (nbPanel.session) {
            // Jupyter 1.0
            nbPanel.session.statusChanged.connect((_, status) => {
                // console.log('SparkMonitor on statusChanged', status);
                if (status === 'starting' || status === 'connected') {
                    // start comm after kernel restarted or switched
                    this.startComm();
                }
            });
        } else {
            // JupyterLab 2.0
            nbPanel.sessionContext.statusChanged.connect((_, status) => {
                // console.log('SparkMonitor on statusChanged', status);
                if (status === 'restarting') {
                    // start comm after kernel restarted
                    this.startComm();
                } else if (status === 'idle') {
                    // clear current cell after cell execution end
                    currentcell.cellExecuteEnd();
                }
            });
            nbPanel.sessionContext.connectionStatusChanged.connect((_, status) => {
                // console.log('SparkMonitor on connectionStatusChanged', status);
                if (status === 'connected') {
                    // start comm after kernel switched
                    this.startComm();
                }
            });
        }
    }

    /**
     * Returns the CellMonitor given a id
     * @param {string} id - The Jupyter id
     * @return {CellMonitor} The CellMonitor object for the cell
     */
    getCellMonitor(id) {
        return this.cellmonitors[id];
    }

    removeCellMonitor(cell) {
        var cellmonitor = this.getCellMonitor(cell.id);
        if (cellmonitor) {
            cellmonitor.removeDisplay();
            this.stopCellMonitor(cell.id);
        }
    }

    /**
     * Start a CellMonitor for a cell.
     * @param {CodeCell} cell - The Jupyter CodeCell instance
     * @return {CellMonitor} The CellMonitor object for the cell
     */
    startCellMonitor(cell) {
        if (this.cellmonitors[cell.id]) {
            this.cellmonitors[cell.id].removeDisplay();
        }
        // console.log(cell)
        for (let cell_widget of this.nbPanel.content.widgets) {
            // console.log(cell_widget)
            if (cell_widget.model.id === cell.id) {
                this.cellmonitors[cell.id] = new CellMonitor(this, cell_widget);
                break;
            }
        }

        return this.cellmonitors[cell.id];
    }

    /**
     * Stop the CellMonitor for a cell.
     * @param {CodeCell} cell - The Jupyter CodeCell instance
     */
    stopCellMonitor(id) {
        if (this.cellmonitors[id]) {
            this.cellmonitors[id].removeDisplay();
            this.cellmonitors[id] = null;
            delete this.cellmonitors[id];
        }
    }

    // -----Functions to show/hide all displays
    /** Toggle all displays. */
    toggleAll() {
        if (this.display_mode === 'hidden')
            this.showAll();
        else if (this.display_mode === 'shown')
            this.hideAll();
    }

    /** Show all displays. */
    showAll() {
        this.display_mode = 'shown';
    }

    /** Hide all displays. */
    hideAll() {
        Object.keys(this.cellmonitors).forEach(id => {
            if (
                Object.prototype.hasOwnProperty.call(this.cellmonitors, id) &&
                this.cellmonitors[id].displayVisible === true) {
                this.cellmonitors[id].removeDisplay();
            }
        });
        this.display_mode = 'hidden';
    }

    // ------Functions to communicate with kernel

    /**
     * Called when comm to kernel is closed.
     * @param {Object} msg - The JSON parsed close message object.
     */
    static onCommClose(msg) {
        console.log('SparkMonitor: Comm Close Message:', msg);
    }

    /**
     * Starts communication with the kernel.
     * Closes any existing communication.
     */
    startComm() {
        console.log('SparkMonitor: Starting Comm with kernel.');
        if (this.nbPanel.session) {
            // Jupyter 1.0
            this.kernel = this.nbPanel.session.kernel;
            this.comm = this.kernel.connectToComm('SynapseSparkMonitor');
        } else {
            // JupyterLab 2.0
            this.kernel = this.nbPanel.sessionContext.session.kernel;
            this.comm = this.kernel.createComm('SynapseSparkMonitor');
        }
        this.comm.open({
            'msgtype': 'FrontEndStart',
            'clienttype': 'JupyterLab',
            'clientid': this.id
        });
        this.comm.onMsg = message => {
            this.handleMessage(message);
        };
        this.comm.onClose = message => {
            SparkMonitor.onCommClose(message);
        };
        console.log('SparkMonitor: Connection with comms established');
    }

    getCellWithGuid(cellGuid, create) {
        var cell = currentcell.getFrontendCell(cellGuid);
        if (cell == null) {
            var cell = currentcell.getRunningCell(this.nbPanel);
            if (cell != null && create) {
                currentcell.addCellPair(cell.model, cellGuid);
            }
        }
        return cell;
    }
    // ------------Message Handling Functions that update the data and delegate to corresponding cell monitors--------------------------------

    /**
     * Called when a Spark job is started.
     * @param {Object} data - The data from the spark listener event.
     */
    onSparkJobStart(data) {
        var cell = this.getCellWithGuid(data.cellGuid, true);
        if (cell == null) {
            console.error('SparkMonitor: Job started with no running cell.');
            return;
        }
        if (cell == false) {
            console.error('SparkMonitor: Job started with cell that already has a another monitor.');
            return;
        }
        // console.log('SparkMonitor: Job Start at cell: ', cell, data);
        var cellmonitor = this.getCellMonitor(cell.id)
        if (!cellmonitor) {
            cellmonitor = this.startCellMonitor(cell);
        }
        this.data['app' + this.app + 'job' + data.jobId] = {
            cell_id: cell.id,
        }
        //These values are set here as previous messages may be missed if reconnecting from a browser reload.
        this.totalCores = data.totalCores;
        this.numExecutors = data.numExecutors;
        if (cellmonitor)
            cellmonitor.onSparkJobStart(data);
    }

    /**
     * Called when a Spark job is ended.
     * @param {Object} data - The data from the spark listener event.
     */
    onSparkJobEnd(data) {
        const id = this.data[`app${this.app}job${data.jobId}`].cell_id;
        if (id) {
            // console.log('SparkMonitor: Job End at cell: ', id, data);
            const cellmonitor = this.getCellMonitor(id);
            if (cellmonitor)
                cellmonitor.onSparkJobEnd(data);
            // currentcell.removeCellPair(data.cellGuid);
            // one cell may correspond to multiple spark jobs
            // removeCellPair should be called after all jobs ended
        } else
            console.error('SparkMonitor:ERROR no cellID for job');
    }

    /**
     * Called when a Spark stage is submitted.
     * @param {Object} data - The data from the spark listener event.
     */
    onSparkStageSubmitted(data) {
        // console.log('SparkMonitor:Stage Submitted', data);
        var cell = this.getCellWithGuid(data.cellGuid, false);
        if (cell == null) {
            console.error('SparkMonitor: Stage started with no running cell.');
            return;
        }
        if (cell == false) {
            console.error('SparkMonitor: Job started with cell that already has a another monitor.');
            return;
        }
        this.data['app' + this.app + 'stage' + data.stageId] = {
            cell_id: cell.id,
        };
        var cellmonitor = this.getCellMonitor(cell.id);
        if (cellmonitor)
            cellmonitor.onSparkStageSubmitted(data);
    }

    /**
     * Called when a Spark stage is completed.
     * @param {Object} data - The data from the spark listener event.
     */
    onSparkStageCompleted(data) {
        // console.log('SparkMonitor:Stage Completed', data);
        const id = this.data[`app${this.app}stage${data.stageId}`].cell_id;
        if (id) {
            const cellmonitor = this.cellmonitors[id];
            if (cellmonitor)
                cellmonitor.onSparkStageCompleted(data);
        } else
            console.error('SparkMonitor:ERROR no cellId for completed stage');
    }

    /**
     * Called when a Spark task is started.
     * @param {Object} data - The data from the spark listener event.
     */
    onSparkTaskStart(data) {
        const id = this.data[`app${this.app}stage${data.stageId}`].cell_id;
        if (id) {
            const cellmonitor = this.getCellMonitor(id);
            if (cellmonitor)
                cellmonitor.onSparkTaskStart(data);
        } else
            console.error('SparkMonitor:ERROR no cellID for task start');
    }

    /**
     * Called when a Spark task is ended.
     * @param {Object} data - The data from the spark listener event.
     */
    onSparkTaskEnd(data) {
        const id = this.data[`app${this.app}stage${data.stageId}`].cell_id;
        if (id) {
            const cellmonitor = this.getCellMonitor(id);
            if (cellmonitor)
                cellmonitor.onSparkTaskEnd(data);
        } else
            console.error('SparkMonitor:ERROR no cellID for task end');
    }

    /**
     * Called when a Spark Application is ended.
     * @param {Object} data - The data from the spark listener event.
     */

    static onSparkApplicationEnd(data) {
        // Nothing to do here.
        // console.log(`SparkMonitor: Spark App ended: ${data}`);
    }

    /**
     * Called when a Spark Application is started.
     * @param {Object} data - The data from the spark listener event.
     */
    onSparkApplicationStart(data) {
        this.appId = data.appId;
        this.appName = data.appName;
        this.appAttemptId = data.appAttemptId;
        this.app = `${this.appId}_${this.appAttemptId}`;
    }

    /**
     * Called when an executor is added.
     * @param {Object} data - The data from the spark listener event.
     */
    onSparkExecutorAdded(data) {
        this.totalCores = data.totalCores;
        this.numExecutors += 1;
        const cell = currentcell.getRunningCell(this.nbPanel)
        if (cell !== null) {
            const cellmonitor = this.getCellMonitor(cell.id);
            if (cellmonitor)
                cellmonitor.onSparkExecutorAdded(data);
        }
    }

    /**
     * Called when a Spark executor is removed.
     * @param {Object} data - The data from the spark listener event.
     */
    onSparkExecutorRemoved(data) {
        this.totalCores = data.totalCores;
        this.numExecutors -= 1;
        const cell = currentcell.getRunningCell(this.nbPanel)
        if (cell !== null) {
            const cellmonitor = this.getCellMonitor(cell.id);
            if (cellmonitor)
                cellmonitor.onSparkExecutorRemoved(data);
        }
    }

    /**
     * Delegates a received message to corresponding function.
     *
     * @param {Object} msg - The JSON parsed message object.
     */
    handleMessage(msg) {
        if (this.display_mode === 'hidden') {
            return;
        }
        if (!msg.content.data.msgtype) {
            console.warn('SparkMonitor: Unknown message');
        }
        if (msg.content.data.msgtype === 'fromlivy') {
            const data = JSON.parse(msg.content.data.msg);
            switch (data.msgtype) {
                case 'sparkJobReceived':
                    var runningCell = currentcell.getRunningCell(this.nbPanel);
                    currentcell.addCellPair(runningCell, data.cellGuid);
                    console.log('SparkMonitor: Spark job received: ' + data.cellGuid);
                    break;
                case 'sparkJobStart':
                    this.onSparkJobStart(data);
                    break;
                case 'sparkJobEnd':
                    this.onSparkJobEnd(data);
                    break;
                case 'sparkStageSubmitted':
                    this.onSparkStageSubmitted(data);
                    break;
                case 'sparkStageCompleted':
                    this.onSparkStageCompleted(data);
                    break;
                case 'sparkTaskStart':
                    this.onSparkTaskStart(data);
                    break;
                case 'sparkTaskEnd':
                    this.onSparkTaskEnd(data);
                    break;
                case 'sparkApplicationStart':
                    this.onSparkApplicationStart(data);
                    break;
                case 'sparkApplicationEnd':
                    SparkMonitor.onSparkApplicationEnd(data);
                    break;
                case 'sparkExecutorAdded':
                    this.onSparkExecutorAdded(data);
                    break;
                case 'sparkExecutorRemoved':
                    this.onSparkExecutorRemoved(data);
                    break;
                case 'sparkMonitorDebug':
                    console.log('SparkMonitor[Debug]:', data.info);
                    break;
                default:
                    break;
            }
        }
    }
}
