/**
Portions Copyright (c) Microsoft Corporation
 */

/**
 * Module to detect the currently running cell.
 *
 * The notebook sends execution requests, and they queue up on the message channel.
 * There is no straight forward way to detect the currently running cell.
 * Here we use a queue to store execution requests and dequeue elements as the kernel becomes idle after the requests
 * @module currentcell
 */

var current_cell = null;

/**Map from backend cell_id to frontend cell */
var cell_map_f2b = {};
var cell_map_b2f = {};

/**Add a pair to the map */
function addCellPair(frontend_cell, backend_cell_id) {
    // console.log('addCellPair', frontend_cell? frontend_cell.id: frontend_cell, '<->', backend_cell_id);
    var frontend_cell_id = frontend_cell.id;
    if (cell_map_f2b.hasOwnProperty(frontend_cell_id)) {
        cell_map_b2f[cell_map_f2b[frontend_cell_id]] = false;
    }
    cell_map_f2b[frontend_cell_id] = backend_cell_id;
    cell_map_b2f[backend_cell_id] = frontend_cell;
}

/**Get a frontend cell using a backend cell_id */
function getFrontendCell(backend_cell_id) {
    var frontend_cell = cell_map_b2f[backend_cell_id];
    if (frontend_cell == undefined) {
        frontend_cell = null;
    }
    // console.log('getFrontendCell', frontend_cell? frontend_cell.id: frontend_cell, '<-', backend_cell_id);
    return frontend_cell;
}

/**Remove a f2b and b2f relation, called when a backend cell finished execution */
function removeCellPair(backend_cell_id) {
    if (cell_map_b2f.hasOwnProperty(backend_cell_id)) {
        var frontend_cell_id = cell_map_b2f[backend_cell_id];
        delete cell_map_f2b[frontend_cell_id];
        delete cell_map_b2f[backend_cell_id];
    }
}

/** Called when an execute.CodeCell event occurs. This means an execute request was sent for the current cell. */
function cellExecuteStarted(event, data) {
    // console.log('SparkMonitor[CurrentCell]: Cell execute called:', data.cell.id, cell_queue);
    current_cell = data.cell;
}

/** Called when the kernel becomes idle. This means that a cell finished executing. */
function cellExecuteEnd() {
    if (current_cell !== null) {
        // console.log('SparkMonitor[CurrentCell]: Cell execute finished:', current_cell.id);
    }

    current_cell = null;
}

/** @return {CodeCell} - The running cell, or null. */
function getRunningCell(nbPanel) {
    // console.log('getRunningCell', current_cell? current_cell.id: current_cell)
    if (current_cell === null) {
        // fall back to activeCell, to support (1) old version of JupyterLab (<1.2.0), (2) client that disabled recordTiming
        console.warn('getRunningCell', 'fallback', nbPanel.content.activeCell.model)
        return nbPanel.content.activeCell.model
    }

    return current_cell
}

/**
 * Called when a cell is deleted
 *
 * @param {event} event - The event object,
 * @param {data} data - data of the event, contains the cell
 */
function cellDeleted(event, data) {
    // console.log('SparkMonitor[CurrentCell]: Cell deleted:', data.cell.id, cell_queue);
    var cell = data.cell;
    if (current_cell !== null && current_cell.id === cell.id) {
        current_cell = null;
    }
}

export default {
    'getRunningCell': getRunningCell,
    'addCellPair': addCellPair,
    'removeCellPair': removeCellPair,
    'getFrontendCell': getFrontendCell,
    'cellExecuteStarted': cellExecuteStarted,
    'cellExecuteEnd': cellExecuteEnd,
    'cellDeleted': cellDeleted,
}
